import asyncio
from concurrent.futures import Future
from concurrent.futures import ProcessPoolExecutor as ProcessPoolExecutorBase
from multiprocessing import cpu_count
from typing import Any

from .compat import EventLoopMixin
from .counters import Statistic


class ProcessPoolStatistic(Statistic):
    processes: int
    done: int
    error: int
    success: int
    submitted: int
    sum_time: float


class ProcessPoolExecutor(ProcessPoolExecutorBase, EventLoopMixin):
    DEFAULT_MAX_WORKERS = max((cpu_count(), 4))

    def __init__(self, max_workers: int = DEFAULT_MAX_WORKERS, **kwargs: Any):
        super().__init__(max_workers=max_workers, **kwargs)
        self._statistic = ProcessPoolStatistic()
        self._statistic.processes = max_workers

    def _statistic_callback(
        self,
        future: Future,
        start_time: float,
        loop: asyncio.AbstractEventLoop,
    ) -> None:
        if future.exception():
            self._statistic.error += 1
        else:
            self._statistic.success += 1
        self._statistic.done += 1
        self._statistic.sum_time += loop.time() - start_time

    def submit(self, *args: Any, **kwargs: Any) -> Future:
        """
        Submit blocking function to the pool
        """
        loop = asyncio.get_running_loop()
        start_time = loop.time()
        future = super().submit(*args, **kwargs)
        self._statistic.submitted += 1
        future.add_done_callback(
            lambda f: self._statistic_callback(f, start_time, loop),
        )
        return future

    def __del__(self) -> None:
        self.shutdown()



import asyncio
import logging
from asyncio import iscoroutinefunction
from dataclasses import dataclass
from typing import Any, Callable, Optional, Set, Tuple, Type

from croniter import croniter

from aiomisc import Service
from aiomisc.cron import CronCallback


log = logging.getLogger(__name__)
ExceptionsType = Tuple[Type[Exception], ...]


@dataclass(frozen=True)
class StoreItem:
    callback: CronCallback
    spec: str
    shield: bool
    suppress_exceptions: ExceptionsType


class CronService(Service):
    _callbacks_storage: Set[StoreItem]

    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)
        self._callbacks_storage = set()

    def register(
        self,
        function: Callable,
        spec: str,
        shield: bool = False,
        suppress_exceptions: ExceptionsType = (),
    ) -> None:
        if not iscoroutinefunction(function):
            raise TypeError("function should be a coroutine %r" % function)
        if not croniter.is_valid(spec):
            raise TypeError("Not valid cron spec %r" % spec)

        self._callbacks_storage.add(
            StoreItem(
                CronCallback(function), spec, shield, suppress_exceptions,
            ),
        )

    async def start(self) -> None:
        for item in self._callbacks_storage:
            item.callback.start(
                spec=item.spec,
                loop=self.loop,
                shield=item.shield,
                suppress_exceptions=item.suppress_exceptions,
            )
        log.info("Cron service %s started", self)

    async def stop(self, exception: Optional[Exception] = None) -> None:
        async def _shutdown(item: StoreItem) -> None:
            await asyncio.gather(item.callback.stop(), return_exceptions=True)

        await asyncio.gather(
            *[_shutdown(store) for store in self._callbacks_storage],
            return_exceptions=True,
        )
        log.info("Cron service %s is stopped", self)



import asyncio
import contextvars
import inspect
import logging
import threading
import time
import warnings
from concurrent.futures import ThreadPoolExecutor as ThreadPoolExecutorBase
from dataclasses import dataclass, field
from functools import partial, wraps
from multiprocessing import cpu_count
from queue import SimpleQueue
from types import MappingProxyType
from typing import (
    Any, Awaitable, Callable, Coroutine, Dict, FrozenSet, Optional, Set, Tuple,
    TypeVar,
)

from ._context_vars import EVENT_LOOP
from .compat import ParamSpec
from .counters import Statistic
from .iterator_wrapper import IteratorWrapper


P = ParamSpec("P")
T = TypeVar("T")
F = TypeVar("F", bound=Callable[..., Any])
log = logging.getLogger(__name__)


def context_partial(
    func: F, *args: Any,
    **kwargs: Any,
) -> Any:
    warnings.warn(
        "context_partial has been deprecated and will be removed",
        DeprecationWarning,
    )
    context = contextvars.copy_context()
    return partial(context.run, func, *args, **kwargs)


class ThreadPoolException(RuntimeError):
    pass


class ThreadPoolStatistic(Statistic):
    threads: int
    done: int
    error: int
    success: int
    submitted: int
    sum_time: float


@dataclass(frozen=True)
class WorkItemBase:
    func: Callable[..., Any]
    statistic: ThreadPoolStatistic
    future: asyncio.Future
    loop: asyncio.AbstractEventLoop
    args: Tuple[Any, ...] = field(default_factory=tuple)
    kwargs: Dict[str, Any] = field(default_factory=dict)
    context: contextvars.Context = field(
        default_factory=contextvars.copy_context,
    )


def _set_workitem_result(
    future: asyncio.Future, result: Optional[Any],
    exception: Optional[BaseException],
) -> Any:
    if future.done():
        return

    if exception:
        future.set_exception(exception)
    else:
        future.set_result(result)


class WorkItem(WorkItemBase):

    def __call__(self, no_return: bool = False) -> None:
        if self.future.done():
            return

        if self.loop.is_closed():
            log.warning("Event loop is closed. Ignoring %r", self.func)
            raise asyncio.CancelledError

        result, exception = None, None
        delta = -time.monotonic()
        try:
            result = self.context.run(
                self.func, *self.args, **self.kwargs,
            )
            self.statistic.success += 1
        except BaseException as e:
            self.statistic.error += 1
            exception = e
        finally:
            delta += time.monotonic()
            self.statistic.sum_time += delta
            self.statistic.done += 1

        if no_return:
            return

        if self.loop.is_closed():
            log.warning(
                "Event loop is closed. Forget execution result for %r",
                self.func,
            )
            raise asyncio.CancelledError

        self.loop.call_soon_threadsafe(
            _set_workitem_result,
            self.future,
            result,
            exception,
        )


class TaskChannelCloseException(RuntimeError):
    pass


class TaskChannel(SimpleQueue):
    closed_event: threading.Event

    def __init__(self) -> None:
        super().__init__()
        self.closed_event = threading.Event()

    def get(self, *args: Any, **kwargs: Any) -> WorkItem:
        if self.closed_event.is_set():
            raise TaskChannelCloseException()

        item: Optional[WorkItem] = super().get(*args, **kwargs)
        if item is None:
            self.put(None)
            raise TaskChannelCloseException()
        return item

    def close(self) -> None:
        self.closed_event.set()
        self.put(None)


def thread_pool_thread_loop(
    tasks: TaskChannel,
    statistic: ThreadPoolStatistic,
    stop_event: threading.Event,
    pool_shutdown_event: threading.Event,
) -> None:
    statistic.threads += 1

    try:
        while not pool_shutdown_event.is_set():
            tasks.get()()
    except (TaskChannelCloseException, asyncio.CancelledError):
        return None
    finally:
        stop_event.set()
        statistic.threads -= 1


class ThreadPoolExecutor(ThreadPoolExecutorBase):
    __slots__ = (
        "__futures", "__pool", "__tasks",
        "__write_lock", "__thread_events",
    )

    DEFAULT_POOL_SIZE = min((max((cpu_count() or 1, 4)), 32))
    SHUTDOWN_TIMEOUT = 10

    def __init__(
        self, max_workers: int = DEFAULT_POOL_SIZE,
        statistic_name: Optional[str] = None,
    ) -> None:
        self.__futures: Set[asyncio.Future[Any]] = set()
        self.__thread_events: Set[threading.Event] = set()
        self.__tasks = TaskChannel()
        self.__write_lock = threading.RLock()
        self.__max_workers = max_workers
        self.__shutdown_event = threading.Event()
        self._statistic = ThreadPoolStatistic(statistic_name)

        threads = set()
        # starting minimum threads as possible
        for idx in range(2 if max_workers > 1 else 1):
            threads.add(self._start_thread(idx))

        self.__pool: FrozenSet[threading.Thread] = frozenset(threads)

    def _start_thread(self, idx: int) -> threading.Thread:
        if self.__shutdown_event.is_set():
            raise RuntimeError("Can not create a thread after shutdown")

        event = threading.Event()
        self.__thread_events.add(event)

        thread_name = f"Thread {idx}"

        if self._statistic.name:
            thread_name += f" from pool {self._statistic.name}"

        thread = threading.Thread(
            target=thread_pool_thread_loop,
            name=thread_name.strip(),
            daemon=True,
            args=(
                self.__tasks,
                self._statistic,
                event,
                self.__shutdown_event,
            ),
        )

        thread.start()
        return thread

    def submit(  # type: ignore
        self, fn: F, *args: Any, **kwargs: Any,
    ) -> asyncio.Future:
        """
        Submit blocking function to the pool
        """
        if fn is None or not callable(fn):
            raise ValueError("First argument must be callable")

        with self.__write_lock:
            if self.__shutdown_event.is_set():
                raise RuntimeError("Pool is shutdown")

            if (
                len(self.__pool) < self.__max_workers and
                len(self.__futures) >= len(self.__pool)
            ):
                self._adjust_thread_count()

            loop = asyncio.get_event_loop()
            future: asyncio.Future = loop.create_future()
            self.__futures.add(future)
            future.add_done_callback(self.__futures.discard)

            self.__tasks.put_nowait(
                WorkItem(
                    func=fn, args=args, kwargs=kwargs, loop=loop,
                    future=future, statistic=self._statistic,
                ),
            )

            self._statistic.submitted += 1
            return future

    # noinspection PyMethodOverriding
    def shutdown(self, wait: bool = True) -> None:  # type: ignore
        with self.__write_lock:
            if self.__shutdown_event.is_set():
                return None

            self.__shutdown_event.set()

            del self.__pool

            futures = self.__futures
            del self.__futures

            thread_events = self.__thread_events
            del self.__thread_events

            self.__tasks.close()

            while futures:
                future = futures.pop()
                if future.done():
                    continue
                future.set_exception(ThreadPoolException("Pool closed"))

            if not wait:
                return None

            start_time = time.monotonic()
            while not all(e.is_set() for e in thread_events):
                time.sleep(0.01)
                if (time.monotonic() - start_time) > self.SHUTDOWN_TIMEOUT:
                    log.warning(
                        "Waiting for shutting down the pool %r "
                        "cancelled due to timeout",
                        self,
                    )
                    return None

    def _adjust_thread_count(self) -> None:
        pool_size = len(self.__pool)
        new_threads_count = (
            min(pool_size * 2, self.__max_workers) - pool_size
        )
        if new_threads_count <= 0:
            return None

        threads = []
        for idx in range(pool_size, pool_size + new_threads_count):
            threads.append(self._start_thread(idx))
        self.__pool = frozenset(list(self.__pool) + threads)

    def __del__(self) -> None:
        self.__tasks.close()


def run_in_executor(
    func: Callable[..., T],
    executor: Optional[ThreadPoolExecutorBase] = None,
    args: Any = (),
    kwargs: Any = MappingProxyType({}),
) -> Awaitable[T]:
    try:
        loop = asyncio.get_running_loop()
        return loop.run_in_executor(
            executor, partial(func, *args, **kwargs),
        )
    except RuntimeError:
        # In case the event loop is not running right now is
        # returning coroutine to avoid DeprecationWarning in Python 3.10
        async def lazy_wrapper() -> T:
            loop = asyncio.get_running_loop()
            return await loop.run_in_executor(
                executor, partial(func, *args, **kwargs),
            )
        return lazy_wrapper()


async def _awaiter(future: asyncio.Future) -> T:
    try:
        result = await future
        return result
    except asyncio.CancelledError as e:
        if not future.done():
            future.set_exception(e)
        raise


def threaded(
    func: Callable[P, T],
) -> Callable[P, Awaitable[T]]:
    if asyncio.iscoroutinefunction(func):
        raise TypeError("Can not wrap coroutine")

    if inspect.isgeneratorfunction(func):
        return threaded_iterable(func)

    @wraps(func)
    def wrap(
        *args: P.args, **kwargs: P.kwargs,
    ) -> Awaitable[T]:
        return run_in_executor(func=func, args=args, kwargs=kwargs)

    return wrap


def run_in_new_thread(
    func: F,
    args: Any = (),
    kwargs: Any = MappingProxyType({}),
    detach: bool = True,
    no_return: bool = False,
    statistic_name: Optional[str] = None,
) -> asyncio.Future:
    loop = asyncio.get_event_loop()
    future = loop.create_future()

    statistic = ThreadPoolStatistic(statistic_name)
    statistic.threads += 1

    thread = threading.Thread(
        target=WorkItem(
            func=func,
            args=args,
            kwargs=kwargs,
            loop=loop,
            future=future,
            statistic=statistic,
        ),
        name=func.__name__,
        kwargs=dict(no_return=no_return),
        daemon=detach,
    )
    statistic.submitted += 1
    loop.call_soon(thread.start)
    return future


def threaded_separate(
    func: F,
    detach: bool = True,
) -> Callable[..., Awaitable[Any]]:
    if isinstance(func, bool):
        return partial(threaded_separate, detach=detach)

    if asyncio.iscoroutinefunction(func):
        raise TypeError("Can not wrap coroutine")

    @wraps(func)
    def wrap(*args: Any, **kwargs: Any) -> Any:
        future = run_in_new_thread(
            func, args=args, kwargs=kwargs, detach=detach,
        )
        return future

    return wrap


def threaded_iterable(
    func: Optional[F] = None,
    max_size: int = 0,
) -> Any:
    if isinstance(func, int):
        return partial(threaded_iterable, max_size=func)
    if func is None:
        return partial(threaded_iterable, max_size=max_size)

    @wraps(func)
    def wrap(*args: Any, **kwargs: Any) -> Any:
        return IteratorWrapper(
            partial(func, *args, **kwargs),
            max_size=max_size,
        )

    return wrap


class IteratorWrapperSeparate(IteratorWrapper):
    def _run(self) -> Any:
        return run_in_new_thread(self._in_thread)


def threaded_iterable_separate(
    func: Optional[F] = None,
    max_size: int = 0,
) -> Any:
    if isinstance(func, int):
        return partial(threaded_iterable_separate, max_size=func)
    if func is None:
        return partial(threaded_iterable_separate, max_size=max_size)

    @wraps(func)
    def wrap(*args: Any, **kwargs: Any) -> Any:
        return IteratorWrapperSeparate(
            partial(func, *args, **kwargs),
            max_size=max_size,
        )

    return wrap


class CoroutineWaiter:
    def __init__(
        self, coroutine: Coroutine[Any, Any, T],
        loop: Optional[asyncio.AbstractEventLoop] = None,
    ):
        self.__coro: Coroutine[Any, Any, T] = coroutine
        self.__loop = loop or EVENT_LOOP.get()
        self.__event = threading.Event()
        self.__result: Optional[T] = None
        self.__exception: Optional[BaseException] = None

    def _on_result(self, task: asyncio.Future) -> None:
        self.__exception = task.exception()
        if self.__exception is None:
            self.__result = task.result()
        self.__event.set()

    def _awaiter(self) -> None:
        task: asyncio.Future = self.__loop.create_task(self.__coro)
        task.add_done_callback(self._on_result)

    def start(self) -> None:
        self.__loop.call_soon_threadsafe(self._awaiter)

    def wait(self) -> Any:
        self.__event.wait()
        if self.__exception is not None:
            raise self.__exception
        return self.__result


def wait_coroutine(
    coro: Coroutine[Any, Any, T],
    loop: Optional[asyncio.AbstractEventLoop] = None,
) -> T:
    waiter = CoroutineWaiter(coro, loop)
    waiter.start()
    return waiter.wait()


def sync_wait_coroutine(
    loop: Optional[asyncio.AbstractEventLoop],
    coro_func: Callable[..., Coroutine[Any, Any, T]],
    *args: Any,
    **kwargs: Any,
) -> T:
    return wait_coroutine(coro_func(*args, **kwargs), loop=loop)


def sync_await(
    func: Callable[..., Awaitable[T]],
    *args: Any,
    **kwargs: Any,
) -> T:
    async def awaiter() -> T:
        return await func(*args, **kwargs)
    return wait_coroutine(awaiter())



import logging
import sys
from argparse import ArgumentParser
from pathlib import Path

from aiomisc_log import LogFormat, LogLevel, basic_config

from . import plugins


executable = Path(sys.executable)
module_name = "aiomisc.plugins"
parser = ArgumentParser(prog=f"{executable.name} -m {module_name}")
parser.add_argument(
    "-q", "-s", "--quiet", "--silent", action="store_true",
    help="Disable logs and just output plugin-list, alias for "
         "--log-level=critical",
)

parser.add_argument(
    "-n", "--no-output", action="store_true",
    help="Disable output plugin-list to the stdout",
)

parser.add_argument(
    "-l", "--log-level", choices=LogLevel.choices(),
    default=LogLevel.default(), help="Logging level",
)

parser.add_argument(
    "-F", "--log-format", choices=LogFormat.choices(),
    default=LogFormat.default(), help="Logging format",
)


def main() -> None:
    arguments = parser.parse_args()

    if arguments.quiet:
        arguments.log_level = LogLevel.critical.name

    basic_config(log_format=arguments.log_format, level=arguments.log_level)
    logging.info("Available %s plugins.", len(plugins))

    for name, plugin in sorted(plugins.items()):
        logging.info("%r - %s", name, getattr(plugin, "__doc__", ""))

        if not arguments.no_output:
            print(name)


if __name__ == "__main__":
    main()



import asyncio
from collections import defaultdict
from typing import Any, DefaultDict, Dict, Optional

from aiomisc.counters import Statistic


_StorageType = DefaultDict[Any, asyncio.Future]
_EventObjectStoreType = Dict[
    asyncio.AbstractEventLoop, "Context",
]


class ContextStatistic(Statistic):
    get: int
    set: int


class Context:
    __slots__ = ("_storage", "_loop", "_statistic")

    _EVENT_OBJECTS = {}     # type: _EventObjectStoreType

    def close(self) -> None:
        self._storage.clear()
        self._EVENT_OBJECTS.pop(self._loop, None)

    def __init__(self, loop: asyncio.AbstractEventLoop):
        self._loop = loop
        self._storage = defaultdict(loop.create_future)  # type: _StorageType
        self._EVENT_OBJECTS[loop] = self
        self._statistic = ContextStatistic()

    def __getitem__(self, item: Any) -> Any:
        self._statistic.get += 1
        return self._storage[item]

    def __setitem__(self, item: Any, value: Any) -> None:
        self._statistic.set += 1
        self._loop.call_soon_threadsafe(self.__setter, item, value)

    def __setter(self, item: Any, value: Any) -> None:
        if self._storage[item].done():
            del self._storage[item]

        self._storage[item].set_result(value)


def get_context(loop: Optional[asyncio.AbstractEventLoop] = None) -> Context:
    loop = loop or asyncio.get_event_loop()

    if loop.is_closed():
        raise RuntimeError("event loop is closed")

    return Context._EVENT_OBJECTS[loop]


__all__ = ("get_context", "Context")



import atexit
import gc
import os
import signal
import sys
from contextlib import suppress
from threading import Event, Lock
from typing import Any, MutableMapping, Optional, Tuple

from aiomisc_log import basic_config

from . import INT_SIGNAL, AddressType, log
from .protocol import FileIOProtocol
from .worker import bad_initializer, worker


PROCESSES: MutableMapping[int, bytes] = {}
STOPPING = Event()
EXIT_LOCK = Lock()


def at_exit() -> None:
    global PROCESSES

    with EXIT_LOCK:
        pids = tuple(PROCESSES.keys())
        PROCESSES.clear()

    for pid in pids:
        log.debug("Terminating PID: %d", pid)
        with suppress(Exception):
            os.kill(pid, signal.SIGTERM)


def handle_interrupt(*_: Any) -> None:
    STOPPING.set()
    raise KeyboardInterrupt


DEFAULT_SIGNAL_HANDLER = signal.getsignal(INT_SIGNAL)


def fork(worker_id: bytes, cookie: bytes, address: AddressType) -> None:
    global PROCESSES

    gc.disable()
    pid = os.fork()
    gc.enable()

    if pid != 0:
        # In master process
        PROCESSES[pid] = worker_id
        log.debug("Started child process PID: %d", pid)
        return

    # In child process
    PROCESSES = {}
    signal.signal(INT_SIGNAL, DEFAULT_SIGNAL_HANDLER)
    worker(address, cookie, worker_id)
    raise SystemExit(0)


def main() -> int:
    global STOPPING

    proto_stdin = FileIOProtocol(sys.stdin.buffer)

    log_level, log_format = proto_stdin.receive()
    basic_config(level=log_level, log_format=log_format)

    address: AddressType = proto_stdin.receive()
    cookie: bytes = proto_stdin.receive()
    worker_ids: Tuple[bytes, ...] = proto_stdin.receive()
    initializer, initializer_args, initializer_kwargs = proto_stdin.receive()
    worker_id: Optional[bytes]

    def run_initializer() -> None:
        # Saving the initializer result and prevent freeing it
        if not initializer:
            return

        # noinspection PyBroadException
        try:
            initializer(*initializer_args, **initializer_kwargs)
        except BaseException as e:
            log.exception(
                "WorkerPool initializer %r has been failed", initializer,
            )
            bad_initializer(address, cookie, worker_ids[0], e)
            raise SystemExit(0)

    sys.stdin.close()
    del proto_stdin

    run_initializer()

    signal.signal(INT_SIGNAL, handle_interrupt)
    atexit.register(at_exit)

    log.debug("Forking %d processes", len(worker_ids))
    for worker_id in worker_ids:
        fork(worker_id, cookie, address)

    log.debug("Waiting workers")
    while not STOPPING.is_set():
        try:
            pid, status = os.wait()
        except KeyboardInterrupt:
            return exit(0)

        log_func = log.debug if status == 0 else log.warning
        log_func("Worker PID: %d exited with status %d", pid, status)

        if status == 0:
            continue
        worker_id = PROCESSES.pop(pid, None)
        if worker_id is None:
            continue
        fork(worker_id, cookie, address)
    return 0



import asyncio
import logging
import socket
from abc import ABC, abstractmethod
from functools import partial
from pathlib import Path
from typing import Any, Optional, Tuple

from ..utils import OptionsType, TimeoutType, awaitable, bind_socket
from .base import SimpleClient, SimpleServer


log = logging.getLogger(__name__)


class TCPServer(SimpleServer):
    PROTO_NAME = "tcp"

    def __init__(
        self, address: Optional[str] = None, port: Optional[int] = None,
        options: OptionsType = (), sock: Optional[socket.socket] = None,
        **kwargs: Any,
    ):
        if not sock:
            if address is None or port is None:
                raise RuntimeError(
                    "You should pass socket instance or "
                    '"address" and "port" couple',
                )

            self.make_socket = partial(
                bind_socket,
                address=address,
                port=port,
                options=options,
            )
        elif not isinstance(sock, socket.socket):
            raise ValueError("sock must be socket instance")
        else:
            self.make_socket = lambda: sock     # type: ignore

        self.socket: Optional[socket.socket] = None

        super().__init__(**kwargs)

    @abstractmethod
    async def handle_client(
        self, reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
    ) -> Any:
        raise NotImplementedError

    async def __client_handler(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter,
    ) -> Any:
        return await awaitable(self.handle_client)(reader, writer)

    def make_client_handler(
        self, reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
    ) -> Any:
        return self.create_task(self.__client_handler(reader, writer))

    async def start(self) -> None:
        self.socket = self.make_socket()
        self.server = await asyncio.start_server(
            self.make_client_handler,
            sock=self.socket,
        )

    async def stop(self, exc: Optional[Exception] = None) -> None:
        await super().stop(exc)

        if self.server:
            await self.server.wait_closed()

    @property
    def __sockname(self) -> Optional[tuple]:
        if self.socket:
            return self.socket.getsockname()
        return None

    @property
    def address(self) -> Optional[str]:
        if self.__sockname:
            return self.__sockname[0]
        return None

    @property
    def port(self) -> Optional[int]:
        if self.__sockname:
            return self.__sockname[1]
        return None


class TCPClient(SimpleClient, ABC):
    PROTO_NAME = "tcp"

    address: str
    port: int
    path: Path

    connect_attempts: int = 5
    connect_retry_timeout: TimeoutType = 3

    def __init__(
        self,
        address: Optional[str] = None,
        port: Optional[int] = None,
        **kwargs: Any,
    ):
        if not (address and port):
            raise RuntimeError('You should pass "address" and "port" couple')
        self.address = address
        self.port = port
        super().__init__(**kwargs)

    async def connect(self) -> Tuple[
        asyncio.StreamReader, asyncio.StreamWriter,
    ]:
        last_error: Optional[Exception] = None

        for _ in range(self.connect_attempts):
            try:
                reader, writer = await self.create_task(
                    asyncio.open_connection(
                        self.address, self.port,
                    ),
                )
                log.info(
                    "Connected to %s://%s:%d",
                    self.PROTO_NAME, self.address, self.port,
                )
                return reader, writer
            except ConnectionError as e:
                last_error = e
                await asyncio.sleep(self.connect_retry_timeout)

        raise last_error or RuntimeError("Unknown exception occurred")

    @abstractmethod
    async def handle_connection(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter,
    ) -> None:
        raise NotImplementedError

    async def start(self) -> None:
        reader, writer = await self.connect()
        self.loop.call_soon(self.start_event.set)
        try:
            await self.create_task(self.handle_connection(reader, writer))
        except Exception:
            log.exception("Error when handle TCP connection")
        else:
            log.info("Client %r communication has been ended", self)


class RobustTCPClient(TCPClient, ABC):
    reconnect_timeout: TimeoutType = 5
    connected: asyncio.Event

    async def start(self) -> None:
        self.connected = asyncio.Event()

        while True:
            try:
                reader, writer = await self.connect()
                self.connected.set()

                if not self.start_event.is_set():
                    self.start_event.set()

                await self.create_task(self.handle_connection(reader, writer))
                self.connected.clear()
            except asyncio.CancelledError:
                raise
            except Exception:
                log.exception(
                    "Error when handle TCP connection, "
                    "reconnecting after %.3f seconds",
                    self.reconnect_timeout,
                )
            finally:
                if not self.connected.is_set():
                    return
                await asyncio.sleep(self.reconnect_timeout)



import logging
from typing import Any


try:
    from logging_journald import JournaldLogHandler  # type: ignore

    def journald_formatter(**_: Any) -> logging.Handler:
        if not JournaldLogHandler.SOCKET_PATH.exists():
            raise FileNotFoundError("JournalD socket doesn't exists")

        handler = JournaldLogHandler()
        formatter = logging.Formatter("%(message)s")
        handler.setFormatter(formatter)
        return handler

except ImportError:
    def journald_formatter(**_: Any) -> logging.Handler:
        raise ImportError(
            "You must install \"logging-journald\" library for use it",
        )



import asyncio
import logging
import socket
import ssl
from abc import ABC, abstractmethod
from dataclasses import dataclass
from functools import partial
from pathlib import Path
from typing import Any, Optional, Tuple, Union

from ..utils import OptionsType, TimeoutType, awaitable, bind_socket
from .base import SimpleServer
from .tcp import RobustTCPClient, TCPClient


PathOrStr = Union[Path, str]
log = logging.getLogger(__name__)


@dataclass(frozen=True)
class SSLOptionsBase:
    cert: Optional[Path]
    key: Optional[Path]
    ca: Optional[Path]
    verify: bool
    require_client_cert: bool
    purpose: ssl.Purpose


class SSLOptions(SSLOptionsBase):
    def __init__(
        self, cert: Optional[PathOrStr], key: Optional[PathOrStr],
        ca: Optional[PathOrStr], verify: bool, require_client_cert: bool,
        purpose: ssl.Purpose,
    ) -> None:
        super().__init__(
            cert=Path(cert) if cert else None,
            key=Path(key) if key else None,
            ca=Path(ca) if ca else None,
            verify=verify,
            require_client_cert=require_client_cert,
            purpose=purpose,
        )

    def create_context(self) -> ssl.SSLContext:
        context = ssl.create_default_context(
            purpose=self.purpose, cafile=self.ca,
        )

        if self.ca and not self.ca.exists():
            raise FileNotFoundError("CA file doesn't exists")

        if self.require_client_cert:
            context.verify_mode = ssl.VerifyMode.CERT_REQUIRED

        if self.key and self.cert:
            context.load_cert_chain(self.cert, self.key)

        if not self.verify:
            context.check_hostname = False

        return context


class TLSServer(SimpleServer):
    PROTO_NAME = "tls"

    def __init__(
        self, *, address: Optional[str] = None, port: Optional[int] = None,
        cert: PathOrStr, key: PathOrStr, ca: Optional[PathOrStr] = None,
        require_client_cert: bool = False, verify: bool = True,
        options: OptionsType = (), sock: Optional[socket.socket] = None,
        **kwargs: Any,
    ):

        self.__ssl_options = SSLOptions(
            cert, key, ca, verify, require_client_cert,
            ssl.Purpose.CLIENT_AUTH,
        )

        if not sock:
            if address is None or port is None:
                raise RuntimeError(
                    "You should pass socket instance or "
                    '"address" and "port" couple',
                )

            self.make_socket = partial(
                bind_socket,
                address=address,
                port=port,
                options=options,
            )
        elif not isinstance(sock, socket.socket):
            raise ValueError("sock must be socket instance")
        else:
            self.make_socket = lambda: sock     # type: ignore

        self.socket: Optional[socket.socket] = None

        super().__init__(**kwargs)

    async def __client_handler(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter,
    ) -> Any:
        return await awaitable(self.handle_client)(reader, writer)

    def make_client_handler(
        self, reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
    ) -> asyncio.Task:
        return self.create_task(self.__client_handler(reader, writer))

    @abstractmethod
    async def handle_client(
        self, reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
    ) -> Any:
        raise NotImplementedError

    async def start(self) -> None:
        ssl_context = await self.loop.run_in_executor(
            None, self.__ssl_options.create_context,
        )

        self.socket = self.make_socket()

        self.server = await asyncio.start_server(
            self.make_client_handler,
            sock=self.socket,
            ssl=ssl_context,
        )

    async def stop(self, exc: Optional[Exception] = None) -> None:
        await super().stop(exc)

        if self.server:
            await self.server.wait_closed()

    @property
    def __sockname(self) -> Optional[tuple]:
        if self.socket:
            return self.socket.getsockname()
        return None

    @property
    def address(self) -> Optional[str]:
        if self.__sockname:
            return self.__sockname[0]
        return None

    @property
    def port(self) -> Optional[int]:
        if self.__sockname:
            return self.__sockname[1]
        return None


class TLSClient(TCPClient, ABC):
    __slots__ = "__ssl_options", "__server_hostname"

    connect_attempts: int = 5
    connect_retry_timeout: TimeoutType = 3

    def __init__(
        self, address: str, port: int, *,
        cert: PathOrStr, key: PathOrStr,
        ca: Optional[PathOrStr] = None, verify: bool = True,
        **kwargs: Any,
    ) -> None:
        self.__ssl_options = SSLOptions(
            cert, key, ca, verify, False, ssl.Purpose.SERVER_AUTH,
        )

        super().__init__(address, port, **kwargs)

    async def connect(self) -> Tuple[
        asyncio.StreamReader, asyncio.StreamWriter,
    ]:
        last_error: Optional[Exception] = None
        for _ in range(self.connect_attempts):
            try:
                ssl_context: ssl.SSLContext = await self.loop.run_in_executor(
                    None, self.__ssl_options.create_context,
                )
                reader, writer = await asyncio.open_connection(
                    self.address, self.port,
                    ssl=ssl_context,
                )
                log.info(
                    "Connected to %s://%s:%d",
                    self.PROTO_NAME, self.address, self.port,
                )
                return reader, writer
            except ConnectionError as e:
                last_error = e
                await asyncio.sleep(self.connect_retry_timeout)

        raise last_error or RuntimeError("Unknown exception occurred")


class RobustTLSClient(RobustTCPClient, ABC):
    __slots__ = "__ssl_options", "__server_hostname"

    connect_attempts: int = 5
    connect_retry_timeout: TimeoutType = 3

    def __init__(
        self, address: str, port: int, *,
        cert: PathOrStr, key: PathOrStr,
        ca: Optional[PathOrStr] = None, verify: bool = True,
        **kwargs: Any,
    ):
        self.__ssl_options = SSLOptions(
            cert, key, ca, verify, False, ssl.Purpose.SERVER_AUTH,
        )

        super().__init__(address, port, **kwargs)

    async def connect(self) -> Tuple[
        asyncio.StreamReader, asyncio.StreamWriter,
    ]:
        last_error: Optional[Exception] = None
        for _ in range(self.connect_attempts):
            try:
                reader, writer = await asyncio.open_connection(
                    self.address, self.port,
                    ssl=await self.loop.run_in_executor(
                        None, self.__ssl_options.create_context,
                    ),
                )
                log.info(
                    "Connected to %s://%s:%d",
                    self.PROTO_NAME, self.address, self.port,
                )
                return reader, writer
            except ConnectionError as e:
                last_error = e
                await asyncio.sleep(self.connect_retry_timeout)

        raise last_error or RuntimeError("Unknown exception occurred")



import asyncio
from abc import ABC, ABCMeta, abstractmethod
from typing import (
    Any, Coroutine, Dict, Generator, Optional, Set, Tuple, TypeVar, Union,
)

from ..context import Context, get_context
from ..utils import cancel_tasks


T = TypeVar("T")
CoroutineType = Union[Coroutine[Any, Any, T], Generator[Any, None, T]]


class ServiceMeta(ABCMeta):
    def __new__(
        cls, name: str, bases: Tuple, namespace: Dict, **kwds: Any,
    ) -> Any:
        instance = super().__new__(
            cls, name, bases, dict(namespace),
        )

        for key in ("__async_required__", "__required__"):
            setattr(instance, key, frozenset(getattr(instance, key, ())))

        check_instance = all(
            asyncio.iscoroutinefunction(getattr(instance, method))
            for method in instance.__async_required__   # type: ignore
        )

        if not check_instance:
            raise TypeError(
                "Following methods must be coroutine functions",
                tuple(
                    f"{name}.{m}"
                    for m in instance.__async_required__    # type: ignore
                ),
            )

        return instance


class Service(metaclass=ServiceMeta):
    __async_required__: Tuple[str, ...] = ("start", "stop")
    __required__: Tuple[str, ...] = ()

    _instance_params: Dict[str, Any]

    def __new__(cls, *args: Any, **kwargs: Any) -> "Service":
        instance = super().__new__(cls)
        instance._instance_params = {}
        return instance

    def __init__(self, **kwargs: Any):
        lost_kw = self.__required__ - kwargs.keys()
        if lost_kw:
            raise AttributeError("Absent attributes", lost_kw)

        self._set_params(**kwargs)
        self.__context: Optional[Context] = None
        self.__start_event: Optional[asyncio.Event] = None

    def __getattr__(self, key: str) -> Any:
        try:
            return self._instance_params[key]
        except KeyError as e:
            raise AttributeError(key) from e

    def __setattr__(self, key: str, value: Any) -> None:
        super(Service, self).__setattr__(key, value)
        if key in self._instance_params:
            self._instance_params[key] = value

    @property
    def start_event(self) -> asyncio.Event:
        if self.__start_event is None:
            raise RuntimeError

        return self.__start_event

    @property
    def context(self) -> Context:
        if self.__context is None:
            self.__context = get_context()
        return self.__context

    def set_loop(self, loop: asyncio.AbstractEventLoop) -> None:
        # noinspection PyAttributeOutsideInit
        self.loop = loop
        self.__start_event = asyncio.Event()

    def _set_params(self, **kwargs: Any) -> None:
        self._instance_params = kwargs

        for name, value in kwargs.items():
            setattr(self, name, value)

    def __getstate__(self) -> Dict[str, Any]:
        return self._instance_params

    def __setstate__(self, state: Dict[str, Any]) -> None:
        self._set_params(**state)

    @abstractmethod
    async def start(self) -> Any:
        raise NotImplementedError

    async def stop(self, exception: Optional[Exception] = None) -> Any:
        pass


class TaskStoreBase(Service, ABC):
    def __init__(self, **kwargs: Any):
        self.tasks: Set[asyncio.Task] = set()
        super().__init__(**kwargs)

    def create_task(self, coro: CoroutineType) -> asyncio.Task:
        task: asyncio.Task = self.loop.create_task(coro)
        self.tasks.add(task)
        task.add_done_callback(self.tasks.discard)
        return task

    async def stop(self, exc: Optional[Exception] = None) -> None:
        await cancel_tasks(self.tasks)


class SimpleServer(TaskStoreBase):
    def __init__(self, **kwargs: Any):
        self.server: Optional[asyncio.AbstractServer] = None
        self.tasks: Set[asyncio.Task] = set()
        super().__init__(**kwargs)

    @abstractmethod
    async def start(self) -> None:
        raise NotImplementedError

    async def stop(self, exc: Optional[Exception] = None) -> None:
        await super().stop(exc)

        if self.server:
            self.server.close()


class SimpleClient(TaskStoreBase):
    @abstractmethod
    async def start(self) -> None:
        raise NotImplementedError



import logging
import os
import sys
from enum import Enum, IntEnum, unique
from typing import Tuple


try:
    from logging_journald import check_journal_stream  # type: ignore
except ImportError:
    def check_journal_stream() -> bool:
        return False

try:
    import rich as _  # noqa
    RICH_INSTALLED = True
except ImportError:
    RICH_INSTALLED = False


@unique
class LogFormat(IntEnum):
    disabled = -1
    stream = 0
    color = 1
    json = 2
    syslog = 3
    plain = 4
    journald = 5
    rich = 6
    rich_tb = 7

    @classmethod
    def choices(cls) -> Tuple[str, ...]:
        return tuple(cls._member_names_)

    @classmethod
    def default(cls) -> str:
        if check_journal_stream():
            return cls.journald.name

        if not os.isatty(sys.stderr.fileno()):
            return cls.plain.name

        if RICH_INSTALLED:
            return cls.rich.name

        return cls.color.name


class LogLevel(IntEnum):
    critical = logging.CRITICAL
    error = logging.ERROR
    warning = logging.WARNING
    info = logging.INFO
    debug = logging.DEBUG
    notset = logging.NOTSET

    @classmethod
    def choices(cls) -> Tuple[str, ...]:
        return tuple(cls._member_names_)

    @classmethod
    def default(cls) -> str:
        return cls.info.name


class DateFormat(Enum):
    color = "%Y-%m-%d %H:%M:%S"
    stream = "[%Y-%m-%d %H:%M:%S]"

    # Optimization: special value ``%s`` date will
    # not formatted just returns record created time
    json = "%s"
    syslog = None
    rich = "[%X]"



import logging
import logging.handlers
import os
import sys
from contextvars import ContextVar
from dataclasses import dataclass
from itertools import chain
from types import TracebackType
from typing import Any, Callable, Iterable, Optional, Type, Union

from .enum import LogFormat, LogLevel
from .formatter import (
    color_formatter, journald_formatter, json_handler, rich_formatter,
)


LOG_LEVEL: ContextVar = ContextVar("LOG_LEVEL", default=logging.INFO)
LOG_FORMAT: ContextVar = ContextVar(
    "LOG_FORMAT", default=LogFormat.default(),
)


DEFAULT_FORMAT = "%(levelname)s:%(name)s:%(message)s"


def create_logging_handler(
    log_format: LogFormat = LogFormat.color,
    date_format: Optional[str] = None, **kwargs: Any,
) -> Optional[logging.Handler]:

    LOG_FORMAT.set(log_format)

    if log_format == LogFormat.disabled:
        return None
    elif log_format == LogFormat.stream:
        handler: logging.Handler = logging.StreamHandler()
        if date_format and date_format is not Ellipsis:
            formatter = logging.Formatter(
                "%(asctime)s " + DEFAULT_FORMAT, datefmt=date_format,
            )
        else:
            formatter = logging.Formatter(DEFAULT_FORMAT)

        handler.setFormatter(formatter)
        return handler
    elif log_format == LogFormat.json:
        return json_handler(date_format=date_format, **kwargs)
    elif log_format == LogFormat.color:
        return color_formatter(date_format=date_format, **kwargs)
    elif log_format == LogFormat.journald:
        return journald_formatter()
    elif log_format == LogFormat.rich:
        return rich_formatter(date_format=date_format, **kwargs)
    elif log_format == LogFormat.rich_tb:
        return rich_formatter(
            date_format=date_format,
            rich_tracebacks=True,
            **kwargs,
        )
    elif log_format == LogFormat.syslog:
        if date_format:
            sys.stderr.write("Can not apply \"date_format\" for syslog\n")
            sys.stderr.flush()

        formatter = logging.Formatter("%(message)s")

        if os.path.exists("/dev/log"):
            handler = logging.handlers.SysLogHandler(address="/dev/log")
        else:
            handler = logging.handlers.SysLogHandler()

        handler.setFormatter(formatter)
        return handler
    elif log_format == LogFormat.plain:
        handler = logging.StreamHandler()
        formatter = logging.Formatter("%(message)s")
        handler.setFormatter(formatter)
        return handler

    raise NotImplementedError


HandlerWrapperType = Callable[[logging.Handler], logging.Handler]


def pass_wrapper(handler: logging.Handler) -> logging.Handler:
    return handler


@dataclass
class UnhandledHookBase:
    logger: logging.Logger
    logger_name: str = "unhandled"
    logger_default_message: str = "Unhandled exception"


class UnhandledHook(UnhandledHookBase):
    def __init__(self, **kwargs: Any) -> None:
        logger = logging.getLogger().getChild(self.logger_name)
        logger.propagate = False
        logger.handlers.clear()
        super().__init__(logger=logger, **kwargs)

    def add_handler(self, handler: logging.Handler) -> None:
        self.logger.handlers.append(handler)


class UnhandledPythonHook(UnhandledHook):
    def __call__(
        self,
        exc_type: Type[BaseException],
        exc_value: BaseException,
        exc_traceback: TracebackType,
    ) -> None:
        self.logger.exception(
            self.logger_default_message,
            exc_info=(exc_type, exc_value, exc_traceback),
        )


def basic_config(
    level: Union[int, str] = LogLevel.info,
    log_format: Union[str, LogFormat] = LogFormat.color,
    handler_wrapper: HandlerWrapperType = pass_wrapper,
    handlers: Iterable[logging.Handler] = (),
    **kwargs: Any,
) -> None:

    if isinstance(level, str):
        level = LogLevel[level]

    logging.basicConfig(handlers=[], level=logging.NOTSET)
    root_logger = logging.getLogger()
    root_logger.handlers.clear()

    if isinstance(log_format, str):
        log_format = LogFormat[log_format]

    unhandled_hook = UnhandledPythonHook()
    sys.excepthook = unhandled_hook     # type: ignore

    logging_handlers = list(
        map(
            handler_wrapper,
            filter(
                None,
                chain(
                    [create_logging_handler(log_format, **kwargs)],
                    handlers,
                ),
            ),
        ),
    )

    LOG_LEVEL.set(level)

    # noinspection PyArgumentList
    logging.basicConfig(
        level=int(level),
        handlers=logging_handlers,
    )

    for handler in logging_handlers:
        unhandled_hook.add_handler(handler)


__all__ = (
    "LogFormat",
    "LogLevel",
    "basic_config",
    "create_logging_handler",
    "LOG_FORMAT",
    "LOG_LEVEL",
)
